// src/Bot.js
const { Client, GatewayIntentBits, Events, Collection, REST, Routes, Partials } = require('discord.js');
const EventEmitter = require('node:events');
const CommandHandler = require('./handlers/CommandHandler.js');
const Parser = require('./parser/Parser.js'); // Import the Parser

class Bot extends EventEmitter {
    constructor(options) {
        super(); // Call the EventEmitter constructor

        if (!options || typeof options !== 'object') {
            throw new Error("Bot options must be provided as an object.");
        }
        if (!options.token || typeof options.token !== 'string') {
            throw new Error("A valid Discord bot token must be provided in options.token.");
        }
        if (!options.intents || !Array.isArray(options.intents)) {
            throw new Error("Intents must be provided as an array in options.intents.");
        }

        // --- Store Options ---
        this.token = options.token;
        this.prefix = options.prefix || "!"; // Default prefix
        const userIntents = options.intents;

        // --- Process Intents ---
        const intentBits = userIntents.map(intentName => {
            const intentBit = GatewayIntentBits[intentName];
            if (intentBit === undefined) {
                console.warn(`[Relay.js Warn] Unknown intent specified: ${intentName}. Skipping.`);
                return null;
            }
            return intentBit;
        }).filter(intent => intent !== null);

        if (intentBits.length === 0 && userIntents.length > 0) {
            console.warn("[Relay.js Warn] No valid intents were resolved from provided names. Discord connection might fail or features may not work.");
        } else if (intentBits.length === 0) {
             console.warn("[Relay.js Warn] No intents provided. Using default intents: Guilds, GuildMessages, MessageContent.");
             // Add defaults if none provided/resolved to ensure basic functionality
              intentBits.push(GatewayIntentBits.Guilds, GatewayIntentBits.GuildMessages, GatewayIntentBits.MessageContent);
        }
         // Ensure GuildMessages is present if MessageContent is, as it's often needed together
         if (intentBits.includes(GatewayIntentBits.MessageContent) && !intentBits.includes(GatewayIntentBits.GuildMessages)) {
             console.warn("[Relay.js Warn] MessageContent intent requires GuildMessages intent. Adding GuildMessages.");
             intentBits.push(GatewayIntentBits.GuildMessages);
         }


        // --- Initialize discord.js Client ---
        // Add partials if needed for events on uncached items (e.g., reactions on old messages)
        this.client = new Client({
            intents: intentBits,
            partials: [Partials.Message, Partials.Channel, Partials.Reaction, Partials.User, Partials.GuildMember] // Common partials
        });

        // --- Parser ---
        this.parser = new Parser(this); // Create the parser instance

        // --- Command Storage ---
        this.prefixCommands = new Collection();
        this.slashCommands = new Collection();
        // --- TODO: Add component handler storage ---
        // this.componentHandlers = new Collection(); // Map customId -> handler logic

        // --- Handlers ---
        this.commandHandler = new CommandHandler(this);

        // --- Setup Event Forwarding ---
        this._setupEvents();
    }

    /**
     * Sets up internal listeners for discord.js client events
     * and triggers Relay.js logic or forwards events.
     * @private
     */
    _setupEvents() {
        this.client.once(Events.ClientReady, async (client) => { // Mark async
            console.log(`[Relay.js] Logged in as ${client.user.tag}`);
            try {
                 await this._registerSlashCommands(); // Register slash commands AFTER client is ready
                 console.log("[Relay.js] Slash command registration process completed.");
             } catch (e) {
                 console.error("[Relay.js] Error during initial slash command registration:", e);
             }
            this.emit('ready', client); // Emit 'ready' event for the user
        });

        // --- Prefix Command Listener ---
        this.client.on(Events.MessageCreate, async (message) => {
            // Ignore DMs, bots, and messages without prefix
            if (!message.guild || message.author.bot || !message.content.startsWith(this.prefix)) {
                return;
            }

            // Basic argument parsing
            const args = message.content.slice(this.prefix.length).trim().split(/ +/);
            const commandName = args.shift()?.toLowerCase();

            if (!commandName) return; // No command name after prefix

            const command = this.prefixCommands.get(commandName);
            if (!command) return; // Command not found

            // console.log(`[Relay.js Debug] Executing prefix command "${commandName}" for ${message.author.tag}`);

            // --- Create Execution Context ---
            const context = {
                message: message,
                author: message.author,
                member: message.member, // Might be null if member left but message exists
                channel: message.channel,
                guild: message.guild,
                client: this.client,
                bot: this,
                variables: {} // Reset variables for each execution
            };

            // --- Execute using Parser ---
            try {
                // Parser handles errors internally and attempts to reply
                await this.parser.parse(command.rawLogic, context);
            } catch (e) {
                // Catch any unexpected errors *outside* the parser's controlled execution
                console.error(`[Relay.js] UNCAUGHT error during prefix command "${commandName}" execution:`, e);
            }
        });

        // --- Interaction (Slash/Component) Listener ---
        this.client.on(Events.InteractionCreate, async (interaction) => {
             let command = null;
             let handler = null; // For components
             let commandName = ''; // For logging
             let customId = ''; // For logging

             // Base context shared by all interaction types
             const context = {
                 interaction: interaction,
                 user: interaction.user,
                 member: interaction.member,
                 channel: interaction.channel,
                 guild: interaction.guild,
                 client: this.client,
                 bot: this,
                 variables: {} // Reset variables for each execution
             };

             try {
                 // --- Slash Command Handling ---
                 if (interaction.isChatInputCommand()) {
                     commandName = interaction.commandName;
                     command = this.slashCommands.get(commandName);
                     if (!command) {
                         console.error(`[Relay.js Error] Received interaction for UNKNOWN slash command: ${commandName}`);
                         // Avoid replying if interaction is potentially gone
                         if (interaction.replied || interaction.deferred) return;
                         try {
                            await interaction.reply({ content: 'Error: This command seems outdated or is not registered correctly.', ephemeral: true });
                         } catch (e) { console.error("Failed to send unknown command reply", e); }
                         return;
                     }

                     // console.log(`[Relay.js Debug] Executing slash command "${commandName}" for ${interaction.user.tag}`);
                     await this.parser.parse(command.rawLogic, context);

                 }
                 // --- Button Handling ---
                 else if (interaction.isButton()) {
                    customId = interaction.customId;
                    // --- TODO: Look up button handler ---
                    // handler = this.componentHandlers.get(customId);
                    if (!handler) {
                         console.warn(`[Relay.js Warn] Received button interaction for unhandled custom ID: ${customId}`);
                         // Optional: Send a generic "Expired or invalid button" message
                          if (!interaction.replied && !interaction.deferred) {
                              try { await interaction.reply({ content: 'This button seems to be invalid or expired.', ephemeral: true }); } catch {}
                          }
                         return; // Stop if no handler found
                    }
                     // console.log(`[Relay.js Debug] Executing button handler "${customId}" for ${interaction.user.tag}`);
                     await this.parser.parse(handler.rawLogic, context);
                 }
                 // --- Select Menu Handling ---
                 else if (interaction.isStringSelectMenu() || interaction.isUserSelectMenu() || interaction.isRoleSelectMenu() || interaction.isChannelSelectMenu() || interaction.isMentionableSelectMenu() ) {
                     customId = interaction.customId;
                     const selectedValues = interaction.values; // Array of selected option values
                     // Add selected values to context for functions like $selectedValues
                     context.selectedValues = selectedValues;
                     // --- TODO: Look up select menu handler ---
                     // handler = this.componentHandlers.get(customId);
                      if (!handler) {
                         console.warn(`[Relay.js Warn] Received select menu interaction for unhandled custom ID: ${customId}`);
                          if (!interaction.replied && !interaction.deferred) {
                             try { await interaction.reply({ content: 'This select menu seems to be invalid or expired.', ephemeral: true }); } catch {}
                          }
                         return;
                     }
                     // console.log(`[Relay.js Debug] Executing select menu handler "${customId}" for ${interaction.user.tag}`);
                     await this.parser.parse(handler.rawLogic, context);

                 }
                 // --- TODO: Add Modal Submit Handling ---
                 // else if (interaction.isModalSubmit()) { ... }

             } catch (e) {
                 // Errors during parsing/execution are mostly caught within parser.parse() and replied there.
                 // This catch is for wholly unexpected errors during the interaction routing itself.
                 console.error(`[Relay.js] UNCAUGHT error during interaction handling (Command: ${commandName || 'N/A'}, CustomID: ${customId || 'N/A'}):`, e);
                 // Attempt a generic reply ONLY if the parser didn't already send one.
                 // This is hard to guarantee, so maybe avoid replying here unless absolutely necessary.
                  // try {
                  //    if (!interaction.replied && !interaction.deferred) {
                  //        await interaction.reply({ content: 'An critical error occurred while processing this interaction.', ephemeral: true });
                  //    }
                  // } catch (replyError) {
                  //     console.error("[Relay.js] Failed to send critical interaction error reply:", replyError);
                  // }
             }
        });

        // Forward other useful events
        this.client.on(Events.Error, (error) => {
            console.error("[Relay.js Client Error]", error);
            this.emit('error', error); // Forward errors to user
        });
        this.client.on(Events.Warn, (info) => {
            console.warn("[Relay.js Client Warn]", info);
            this.emit('warn', info); // Forward warnings
        });
        // Optional: Add debug listener
        // this.client.on(Events.Debug, (info) => {
        //     console.debug("[Relay.js Client Debug]", info);
        //     this.emit('debug', info);
        // });
    }

    /**
     * Loads command definition files from a specified directory.
     * @param {string} directoryPath - The absolute path to the commands directory.
     */
    loadCommands(directoryPath) {
        console.log(`[Relay.js] Attempting to load command definitions from: ${directoryPath}`);
        try {
            const loaded = this.commandHandler.loadCommandsFromDir(directoryPath);

            // Clear existing commands before loading new ones (allows for reloading)
             this.prefixCommands.clear();
             this.slashCommands.clear();
             // --- TODO: Clear component handlers ---
             // this.componentHandlers.clear();

            loaded.prefixCommands.forEach((cmd, name) => this.prefixCommands.set(name, cmd));
            loaded.slashCommands.forEach((cmd, name) => this.slashCommands.set(name, cmd));
             // --- TODO: Process and store loaded Button/Select handlers ---
             // loaded.componentHandlers.forEach((handler, id) => this.componentHandlers.set(id, handler));

            console.log(`[Relay.js] Loaded ${this.prefixCommands.size} prefix commands.`);
            console.log(`[Relay.js] Loaded ${this.slashCommands.size} slash commands (pending registration).`);
            // console.log(`[Relay.js] Loaded ${this.componentHandlers.size} component handlers.`);

             // If client is already ready, attempt re-registration. Otherwise handled by 'ready' event.
             // Debounce or delay this if loading happens frequently?
             if (this.client.isReady()) {
                 this._registerSlashCommands().catch(e => {
                     console.error("[Relay.js] Error during command reload registration:", e);
                 });
             }

        } catch (error) {
            console.error(`[Relay.js] Failed to load commands from ${directoryPath}:`, error);
            // Decide if the bot should stop or continue with potentially old commands
            // throw error; // Re-throw to potentially stop the bot startup/reload
        }
    }

     /**
     * Registers loaded slash commands with Discord API.
     * Should be called after client is ready.
     * @private
     */
     async _registerSlashCommands() {
         if (this.slashCommands.size === 0) {
             console.log("[Relay.js] No slash commands loaded to register.");
             // If you want to CLEAR existing commands when none are loaded:
             // const rest = new REST({ version: '10' }).setToken(this.token);
             // const clientId = this.client.user?.id;
             // if (clientId) {
             //    try {
             //        await rest.put(Routes.applicationCommands(clientId), { body: [] });
             //        console.log("[Relay.js] Cleared existing global application commands.");
             //    } catch (e) { console.error("[Relay.js] Failed to clear application commands:", e); }
             // }
             return;
         }

        const rest = new REST({ version: '10' }).setToken(this.token);

        // Prepare command data in the format Discord API expects
         const commandData = this.slashCommands.map(cmd => {
            if (!cmd.name || !cmd.description) {
                 console.warn(`[Relay.js Warn] Skipping slash command from file ${cmd.filePath} due to missing name or description.`);
                 return null; // Skip invalid command structures
            }
             return {
                 name: cmd.name,
                 description: cmd.description,
                 options: cmd.options || [], // Ensure options array exists and is valid
             };
         }).filter(cmd => cmd !== null); // Filter out any skipped commands

         if (commandData.length === 0) {
             console.log("[Relay.js] No valid slash command data to register.");
             return;
         }


         try {
            const clientId = this.client.user?.id;
            if (!clientId) {
                // Should not happen if called after 'ready', but check anyway
                console.error("[Relay.js Error] Cannot register slash commands: Client ID is unavailable.");
                return;
            }

             console.log(`[Relay.js] Registering ${commandData.length} application (/) commands globally...`);
             // Use Routes.applicationCommands for global commands
             const data = await rest.put(
                 Routes.applicationCommands(clientId),
                 { body: commandData },
             );
             console.log(`[Relay.js] Successfully registered ${data.length} application commands.`);

             // --- TODO: Implement Guild-specific command registration ---
             // This would involve:
             // 1. A way to define guild IDs in .rl files (e.g., GUILD ID: 123...)
             // 2. Grouping commands by guild ID during loading.
             // 3. Iterating through each guild group and calling:
             //    await rest.put(Routes.applicationGuildCommands(clientId, guildId), { body: guildCommandData });

         } catch (error) {
             console.error("[Relay.js] Failed to register application commands:", error);
             // Log more details from the error if available
             if (error.response?.data?.errors) {
                 console.error("[Relay.js] Discord API Errors:", JSON.stringify(error.response.data.errors, null, 2));
             } else if (error.request) {
                 console.error("[Relay.js] Request Data:", error.request.data);
             }
         }
     }

    /**
     * Logs the bot into Discord.
     * @returns {Promise<string>} The token used to log in.
     */
    login() {
        if (!this.client) throw new Error("Client not initialized.");
        console.log("[Relay.js] Logging in...");
        return this.client.login(this.token); // Returns promise from discord.js
    }

    /**
     * Destroys the discord.js client connection.
     */
    destroy() {
        if (!this.client) return Promise.resolve(); // Nothing to destroy
        console.log("[Relay.js] Destroying client connection...");
        return this.client.destroy(); // Returns promise
    }
}

module.exports = Bot; // Export the Bot class