// src/handlers/CommandHandler.js
const fs = require('node:fs');
const path = require('node:path');
const discord = require('discord.js'); // Needed for ApplicationCommandOptionType mapping

class CommandHandler {
    constructor(botInstance) {
        this.bot = botInstance; // Store reference to the main Bot if needed later
    }

    /**
     * Reads a directory recursively, finds .rl files, and attempts to parse them.
     * @param {string} dirPath - The path to the directory to load commands from.
     * @returns {{prefixCommands: Map<string, object>, slashCommands: Map<string, object>}} Object containing maps of loaded commands.
     */
    loadCommandsFromDir(dirPath) {

        // Helper function to parse a single file's content
        const parseFile = (filePath) => {
            try {
                const fileContent = fs.readFileSync(filePath, 'utf-8');
                return this._parseFileContent(fileContent, filePath);
            } catch (error) {
                 console.error(`[Relay.js] Error reading or processing file ${filePath}:`, error);
                 return []; // Return empty array on file error
            }
        };

        // Recursive function to read directories
        const readDirRecursive = (currentDirPath) => {
             const prefixCmds = new Map();
             const slashCmds = new Map();
             // --- TODO: Add componentHandlers Map ---
             // const componentHandlers = new Map();

             try {
                 const items = fs.readdirSync(currentDirPath);
                 for (const item of items) {
                     const fullPath = path.join(currentDirPath, item);
                     let stat;
                     try {
                         stat = fs.statSync(fullPath);
                     } catch (e) {
                         console.error(`[Relay.js] Error getting stats for ${fullPath}:`, e);
                         continue; // Skip this item if stats fail
                     }


                     if (stat.isDirectory()) {
                         const subResult = readDirRecursive(fullPath);
                         subResult.prefixCommands.forEach((cmd, name) => prefixCmds.set(name, cmd));
                         subResult.slashCommands.forEach((cmd, name) => slashCmds.set(name, cmd));
                         // --- TODO: Merge componentHandlers ---
                         // subResult.componentHandlers.forEach((handler, id) => componentHandlers.set(id, handler));
                     } else if (item.endsWith('.rl')) {
                         const commandsInFile = parseFile(fullPath);
                         commandsInFile.forEach(cmd => {
                             if (cmd.type === 'prefix') {
                                 if (prefixCmds.has(cmd.name)) {
                                     console.warn(`[Relay.js Warn] Duplicate prefix command name "${cmd.name}" found in ${fullPath}. Overwriting.`);
                                 }
                                 prefixCmds.set(cmd.name, cmd);
                             } else if (cmd.type === 'slash') {
                                 if (slashCmds.has(cmd.name)) {
                                     console.warn(`[Relay.js Warn] Duplicate slash command name "${cmd.name}" found in ${fullPath}. Overwriting.`);
                                 }
                                 slashCmds.set(cmd.name, cmd);
                             }
                             // --- TODO: Handle component handlers ---
                             // else if (cmd.type === 'button' || cmd.type === 'select') {
                             //     if (componentHandlers.has(cmd.customId)) {
                             //        console.warn(`[Relay.js Warn] Duplicate component custom ID "${cmd.customId}" found in ${fullPath}. Overwriting.`);
                             //     }
                             //     componentHandlers.set(cmd.customId, cmd);
                             // }
                         });
                     }
                 }
             } catch (error) {
                if (error.code === 'ENOENT') {
                     console.error(`[Relay.js] Command directory segment not found: ${currentDirPath}`);
                } else {
                     console.error(`[Relay.js] Error reading command directory segment ${currentDirPath}:`, error);
                }
                // Depending on desired robustness, might want to throw or just return empty maps
             }
             // --- TODO: Return componentHandlers ---
             return { prefixCommands: prefixCmds, slashCommands: slashCmds /*, componentHandlers*/ };
        };

         return readDirRecursive(dirPath);
    }


    /**
     * Parses the raw content of an .rl file to extract command definitions.
     * @param {string} content - The raw text content of the .rl file.
     * @param {string} filePath - The path of the file being parsed (for logging).
     * @returns {Array<object>} An array of parsed command/handler objects.
     * @private
     */
    _parseFileContent(content, filePath) {
        const definitions = [];
        const prefixCmdRegex = /COMMAND\s+([\w-]+):([\s\S]*?)END COMMAND/gi;
        const slashCmdRegex = /SLASH COMMAND\s+([\w-]+):([\s\S]*?)END SLASH/gi;
        // --- TODO: Add Regex for BUTTON and SELECT ---
        // const buttonRegex = /BUTTON\s+([\w-]+):([\s\S]*?)END BUTTON/gi;
        // const selectRegex = /SELECT\s+([\w-]+):([\s\S]*?)END SELECT/gi;

        let match;

        // Find Prefix Commands
        while ((match = prefixCmdRegex.exec(content)) !== null) {
            const name = match[1].trim();
            const rawLogic = match[2].trim();
            if (name && rawLogic) {
                // console.log(`[Relay.js Debug] Found Prefix Command: "${name}" in ${filePath}`);
                definitions.push({
                    type: 'prefix',
                    name: name,
                    rawLogic: rawLogic,
                    filePath: filePath
                });
            }
        }

        // Find Slash Commands
        while ((match = slashCmdRegex.exec(content)) !== null) {
            const name = match[1].trim();
            const blockContent = match[2].trim(); // Content between SLASH COMMAND name: and END SLASH

            if (name && blockContent) {
                const descriptionMatch = blockContent.match(/^description:\s*(.*)/im); // Match description at start of line
                const optionsMatch = blockContent.match(/^options:\s*([\s\S]*)/im); // Match options at start of line

                // --- Extract Description ---
                const description = descriptionMatch ? descriptionMatch[1].trim() : 'No description provided.';

                // --- Extract Options ---
                const options = [];
                let optionsBlockRawContent = ''; // Keep track of the raw options block text
                if (optionsMatch && optionsMatch[1]) {
                    // Try to isolate the options block content before other definitions start
                    optionsBlockRawContent = optionsMatch[1].split(/^\s*(description:|logic:|action:)/im)[0].trim();
                    const optionLines = optionsBlockRawContent.split('\n');

                    for (const line of optionLines) {
                        const trimmedLine = line.trim();
                        if (!trimmedLine || trimmedLine.startsWith('//') || trimmedLine.startsWith('#')) continue; // Skip empty/commented lines

                        const parts = trimmedLine.split(';');
                        const typeAndName = parts[0]?.split(':');

                        if (typeAndName?.length === 2 && parts.length >= 2) {
                            const optionTypeStr = typeAndName[0].trim().toLowerCase();
                            const optionName = typeAndName[1].trim().toLowerCase();
                            const optionDesc = parts[1].trim();
                            const optionRequired = parts[2]?.trim().toLowerCase() === 'true';

                            const typeMapping = { // Map to discord.js enum values
                                string: discord.ApplicationCommandOptionType.String, text: discord.ApplicationCommandOptionType.String,
                                integer: discord.ApplicationCommandOptionType.Integer, int: discord.ApplicationCommandOptionType.Integer,
                                number: discord.ApplicationCommandOptionType.Number,
                                boolean: discord.ApplicationCommandOptionType.Boolean, bool: discord.ApplicationCommandOptionType.Boolean,
                                user: discord.ApplicationCommandOptionType.User, member: discord.ApplicationCommandOptionType.User,
                                channel: discord.ApplicationCommandOptionType.Channel,
                                role: discord.ApplicationCommandOptionType.Role,
                                mentionable: discord.ApplicationCommandOptionType.Mentionable,
                                attachment: discord.ApplicationCommandOptionType.Attachment, file: discord.ApplicationCommandOptionType.Attachment
                            };
                            const optionType = typeMapping[optionTypeStr];

                            if (optionType === undefined) {
                                console.warn(`[Relay.js Warn] Unknown slash command option type "${optionTypeStr}" in ${filePath} for command "${name}". Defaulting to String.`);
                                options.push({ type: discord.ApplicationCommandOptionType.String, name: optionName, description: optionDesc, required: optionRequired });
                            } else {
                                options.push({ type: optionType, name: optionName, description: optionDesc, required: optionRequired });
                            }
                        } else {
                            console.warn(`[Relay.js Warn] Malformed option line in ${filePath} for slash command "${name}": ${line}`);
                        }
                    }
                } // End options parsing

                // --- Extract Raw Logic ---
                // Start with the full block content
                let logicContent = blockContent;

                // Find the end index of the description definition if it exists
                let lastDefinitionEnd = 0;
                if (descriptionMatch) {
                    const descEndIndex = blockContent.indexOf(descriptionMatch[0]) + descriptionMatch[0].length;
                    lastDefinitionEnd = Math.max(lastDefinitionEnd, descEndIndex);
                }

                // Find the end index of the options definition block if it exists
                if (optionsMatch) {
                    // Need to find the end of the *actual* options block content, not just the `options:` line
                     const optionsEndIndex = blockContent.indexOf(optionsBlockRawContent, optionsMatch.index) + optionsBlockRawContent.length;
                    lastDefinitionEnd = Math.max(lastDefinitionEnd, optionsEndIndex);
                }

                // The raw logic is everything after the last definition ends
                const rawLogic = blockContent.substring(lastDefinitionEnd).trim();

                // --- Create Command Data ---
                const commandData = {
                    type: 'slash',
                    name: name,
                    description: description,
                    options: options, // Use the parsed options array
                    rawLogic: rawLogic,
                    filePath: filePath
                };
                definitions.push(commandData);
            }
        } // End Slash Command Loop


        // --- TODO: Parse BUTTON definitions ---
        // --- TODO: Parse SELECT definitions ---
        // --- TODO: Parse EVENT definitions ---

        if (definitions.length === 0) {
             // console.log(`[Relay.js Debug] No definitions found in ${filePath}.`);
        }

        return definitions;
    }
}

module.exports = CommandHandler;