// src/parser/FunctionManager.js
const discord = require('discord.js'); // Need access to discord.js types and utilities

class FunctionManager {
    constructor(botInstance, parserInstance) {
        this.bot = botInstance;
        this.parser = parserInstance; // To potentially call evaluate for things like $eval
        this.functions = new Map();
        this._registerFunctions();
    }

    /**
     * Executes a registered Relay function.
     * @param {string} name - The name of the function (e.g., 'reply', 'sendMessage'). Case-insensitive.
     * @param {Array<any>} args - The *already evaluated* arguments passed to the function.
     * @param {object} context - The execution context.
     * @returns {Promise<any>} The result of the function execution.
     */
    async execute(name, args, context) {
        const funcData = this.functions.get(name.toLowerCase()); // Case-insensitive lookup

        if (!funcData) {
             // It wasn't a registered function, maybe it's a variable from $let?
             const letValue = context.variables?.[name.toLowerCase()]; // Check stored variables (case-insensitive)
             if (letValue !== undefined) {
                 return letValue;
             }
            // NOTE: Context variables ($authorID etc.) are now primarily handled by
            // substituteContextVariables in Parser.js before evaluation.
            // The _getContextVariable lookup here becomes a fallback.
             const contextVar = this._getContextVariable(name, context);
             if (contextVar !== undefined) {
                 return contextVar;
             }

            throw new Error(`Parser Error: Unknown function or variable "$${name}"`);
        }

        // --- Optional: Argument Count Checking ---
        if (args.length < funcData.minArgs || (funcData.maxArgs !== Infinity && args.length > funcData.maxArgs)) {
           throw new Error(`Parser Error: Invalid argument count for $${name}. Expected ${funcData.minArgs}${funcData.maxArgs === funcData.minArgs ? '' : funcData.maxArgs === Infinity ? '+' : '-' + funcData.maxArgs}, got ${args.length}.`);
        }

        try {
            // Pass evaluated args directly to the function's execute method
            return await funcData.execute(args, context);
        } catch (error) {
            // Log the specific function error and re-throw to be caught by the main parser loop
            console.error(`[Relay.js Func Error] Error executing $${name}:`, error);
            // Make error message more specific
            throw new Error(`Error in $${name}: ${error.message}`);
        }
    }

     /**
     * Gets the value of built-in context variables like $authorID, $channelName etc.
     * Primarily used as a fallback or for direct substitution now.
     * @param {string} name - The variable name (lowercase).
     * @param {object} context - The execution context.
     * @returns {string | number | boolean | undefined} The value or undefined if not found.
     * @private
     */
     _getContextVariable(name, context) {
         const msg = context.message;
         const interaction = context.interaction;
         const user = interaction?.user || msg?.author;
         const member = interaction?.member || msg?.member;
         const channel = interaction?.channel || msg?.channel;
         const guild = interaction?.guild || msg?.guild;
         const interactionOptions = context.interaction?.options; // For slash command options

         // Ensure names are lowercase for case-insensitive matching
         switch (name.toLowerCase()) {
             // Message Context
             case 'messagecontent': return msg?.content;
             case 'messageid': return msg?.id;
             case 'args': return msg?.content?.split(' ').slice(1).join(' '); // Raw args string
             case 'argscount': return msg?.content?.split(' ').slice(1).filter(Boolean).length ?? 0; // Count actual args
             // Interaction Context
             case 'interactionid': return interaction?.id;
             case 'customid': return interaction?.customId; // For components
             case 'commandname': return interaction?.commandName;
             // User/Member Context
             case 'authorid': case 'userid': return user?.id;
             case 'authortag': case 'usertag': return user?.tag;
             case 'authormention': case 'usermention': return user?.toString();
             case 'authorname': case 'username': return user?.username;
             case 'useravatar': return user?.displayAvatarURL({ dynamic: true, size: 1024 });
             case 'memberjoineddate': return member?.joinedTimestamp;
             case 'accountcreationdate': return user?.createdTimestamp;
             case 'memberpermissions': return member?.permissions.toArray().join('; '); // Semicolon separated
             case 'usertimeout': return member?.communicationDisabledUntilTimestamp; // Timestamp or null
             // Channel Context
             case 'channelid': return channel?.id;
             case 'channelname': return channel?.name;
             case 'channeltopic': return channel?.topic;
             case 'channelmention': return channel?.toString();
             case 'channeltype': return channel ? discord.ChannelType[channel.type] : undefined; // String name of type
             // Guild Context
             case 'guildid': return guild?.id;
             case 'guildname': return guild?.name;
             case 'guildicon': return guild?.iconURL({ dynamic: true, size: 1024 });
             case 'membercount': return guild?.memberCount;
             case 'ownerid': return guild?.ownerId;
             // Bot Context
             case 'botid': return this.bot.client.user?.id;
             case 'bottag': return this.bot.client.user?.tag;
             case 'botmention': return this.bot.client.user?.toString();
             case 'botping': return this.bot.client.ws.ping;
             // Utility
             case 'timestamp': return Date.now();
             // Slash Command Options (dynamic based on option name)
             default:
                if (name.toLowerCase().startsWith('option_') && interactionOptions) {
                    const optionName = name.substring('option_'.length);
                    return interactionOptions.get(optionName)?.value; // Get the value directly
                }
                if (name.toLowerCase().startsWith('userid_') && interactionOptions) { // Specific for user options
                    const optionName = name.substring('userid_'.length);
                    return interactionOptions.getUser(optionName)?.id;
                }
                 if (name.toLowerCase().startsWith('roleid_') && interactionOptions) { // Specific for role options
                    const optionName = name.substring('roleid_'.length);
                    return interactionOptions.getRole(optionName)?.id;
                }
                 if (name.toLowerCase().startsWith('channelid_') && interactionOptions) { // Specific for channel options
                    const optionName = name.substring('channelid_'.length);
                    return interactionOptions.getChannel(optionName)?.id;
                }
                // Check let variables as final fallback (though Parser checks first now)
                return context.variables?.[name.toLowerCase()];
         }
     }

    /**
     * Registers all the built-in $functions.
     * @private
     */
    _registerFunctions() {
        const register = (name, executeFunc, minArgs = 0, maxArgs = Infinity) => {
            // Ensure name is lowercase for consistent lookup
            this.functions.set(name.toLowerCase(), { execute: executeFunc, minArgs, maxArgs });
        };

        // --- Basic Output ---
        register('reply', async (args, context) => {
            const content = String(args[0] ?? '');
            // Args[1] could be boolean for ephemeral, or object for embed/components later
            const ephemeral = String(args[1]).toLowerCase() === 'true' || String(args[1]).toLowerCase() === 'ephemeral';

            if (!content && content !== '0') throw new Error("Cannot send an empty reply."); // Allow '0'

            const options = { content: content, ephemeral: ephemeral, allowedMentions: { repliedUser: false } }; // Default no ping

            if (context.message) {
                return (await context.message.reply(options)).id; // Return message ID
            } else if (context.interaction) {
                if (context.interaction.replied || context.interaction.deferred) {
                    return (await context.interaction.followUp(options)).id;
                } else {
                    // Interaction reply returns the interaction itself or null/void, not the message easily
                    await context.interaction.reply(options);
                    // Fetch reply to return message ID if possible (might fail for ephemeral?)
                     try { return (await context.interaction.fetchReply()).id; } catch { return null; }
                }
            } else {
                throw new Error("$reply requires message or interaction context.");
            }
        }, 1, 2); // Allow content only, or content + ephemeral flag

        register('sendMessage', async (args, context) => {
            const channelId = String(args[0]);
            const content = String(args[1] ?? '');
            // Args[2] could be message ID to reply to, or options object later
            if (!channelId || (!content && content !== '0')) throw new Error("Usage: $sendMessage[channelID;content]");

            const channel = await this.bot.client.channels.fetch(channelId).catch(() => null);
            if (!channel || !channel.isTextBased()) throw new Error(`Invalid or non-text channel ID: ${channelId}`);

            try {
                return (await channel.send(content)).id; // Return message ID
            } catch (e) {
                throw new Error(`Failed to send message to channel ${channelId}: ${e.message}`);
            }
        }, 2, 2);

        register('log', (args, context) => { // Synchronous
            const message = args.map(String).join(' '); // Join args into a string
            console.log("[Relay.js Log]", message);
            return ''; // Log doesn't return anything to the script
        }, 1, Infinity);


        // --- Control Flow ---
        register('if', async (args, context) => {
             if (args.length < 2 || args.length > 3) throw new Error("Usage: $if[condition;then;else?]");
             // Arguments are already evaluated by the parser before reaching here
             const condition = args[0];
             const thenBranch = args[1];
             const elseBranch = args[2]; // Will be undefined if not provided

             // Evaluate condition loosely (e.g., "false", 0, "", null, undefined are falsy)
             const isTruthy = condition !== 'false' && condition !== '0' && condition !== '' && condition !== false && condition !== 0 && condition != null;

             if (isTruthy) {
                 // Since args are pre-evaluated, just return the 'then' result
                 return thenBranch;
             } else if (elseBranch !== undefined) {
                 // Return the 'else' result
                 return elseBranch;
             }
             return ''; // No else branch, condition false, return empty string
         }, 2, 3);

         register('stop', (args, context) => { // Synchronous
            context.stopExecution = true;
            return '';
         }, 0, 0);

         register('stopIf', (args, context) => { // Synchronous
             if (args.length !== 1) throw new Error("Usage: $stopIf[condition]");
             const condition = args[0]; // Already evaluated
             const isTruthy = condition !== 'false' && condition !== '0' && condition !== '' && condition !== false && condition !== 0 && condition != null;
             if (isTruthy) {
                 context.stopExecution = true;
             }
             return '';
         }, 1, 1);


        // --- Variables ---
        register('let', (args, context) => { // Synchronous
            if (args.length !== 2) throw new Error("Usage: $let[variableName;value]");
            const varName = String(args[0]).toLowerCase(); // Store keys lowercase for consistency
            const value = args[1]; // Value is already evaluated
            if (!varName.match(/^[\w-]+$/)) throw new Error("Invalid variable name. Use letters, numbers, underscore, hyphen.");
            if (!context.variables) context.variables = {};
            context.variables[varName] = value;
            // console.log(`[Relay.js Debug] Set variable '${varName}' =`, value); // Debugging
            return ''; // $let doesn't output anything
        }, 2, 2);

        register('get', (args, context) => { // Synchronous
             if (args.length !== 1) throw new Error("Usage: $get[variableName]");
             const varName = String(args[0]).toLowerCase();
             const value = context.variables?.[varName];
             // console.log(`[Relay.js Debug] Get variable '${varName}' =`, value); // Debugging
             return value ?? ''; // Return empty string if not found (or null/undefined?)
         }, 1, 1);


        // --- Utilities ---
         register('ping', (args, context) => { // Synchronous access to cached ping
             return this.bot.client.ws.ping;
         }, 0, 0);

         register('args', (args, context) => { // Get raw arguments string for prefix commands
             return context.message?.content?.split(' ').slice(1).join(' ') ?? '';
         }, 0, 0);

         register('arg', (args, context) => { // Get specific argument by index
             const index = parseInt(args[0], 10);
             if (isNaN(index) || index < 1) throw new Error("Usage: $arg[index] - index must be 1 or higher.");
             const messageArgs = context.message?.content?.split(' ').slice(1);
             return messageArgs?.[index - 1] ?? ''; // Return specific arg or empty string
         }, 1, 1);

        register('isEmpty', (args) => { // Synchronous
            const value = args[0];
            return (value === null || value === undefined || String(value).trim() === '');
        }, 1, 1);

         register('textContains', (args) => { // Synchronous
            if (args.length !== 2) throw new Error("Usage: $textContains[text;substring]");
            return String(args[0]).includes(String(args[1]));
        }, 2, 2);

         // Simple check based on message author flag
         register('ignoreBots', (args, context) => { // Synchronous
             if (context.message?.author?.bot) {
                 context.stopExecution = true;
             }
             return '';
         }, 0, 0);

         // Check permissions for the command user (message or interaction)
         register('onlyPerms', async (args, context) => { // Async because reply might be needed
            if (args.length < 1) throw new Error("Usage: $onlyPerms[perm1;perm2;...;optionalReplyMessage]");

             const member = context.interaction?.member || context.message?.member;
             if (!member) {
                 // Cannot check perms if no member context (e.g., DM) - stop execution
                 context.stopExecution = true;
                 // Try to send a reply message if provided as the last arg
                 const potentialReply = args.length > 1 ? args[args.length - 1] : "Permissions check failed (no server context).";
                 if (typeof potentialReply === 'string' && !discord.PermissionsBitField.Flags[potentialReply]) {
                    await this.parser.sendErrorReply(new Error(potentialReply), context);
                 } else {
                     // If last arg is a valid perm, use default message
                     await this.parser.sendErrorReply(new Error("Permissions check failed (no server context)."), context);
                 }
                 return ''; // Stop execution already flagged
             }

             // Determine required perms and optional reply message
             let requiredPerms = [];
             let replyMessage = "You do not have the required permissions!"; // Default reply
             let hasReplyArg = false;
             if (args.length > 1 && typeof args[args.length - 1] === 'string' && !discord.PermissionsBitField.Flags[args[args.length - 1]]) {
                 // If the last argument is a string and NOT a valid permission name, treat it as the reply message
                 requiredPerms = args.slice(0, -1);
                 replyMessage = args[args.length - 1];
                 hasReplyArg = true;
             } else {
                 // Otherwise, all arguments are treated as permissions
                 requiredPerms = args;
             }

             const missingPerms = [];
             for (const permName of requiredPerms) {
                if (typeof permName !== 'string') continue; // Skip non-string args just in case
                 const permBit = discord.PermissionsBitField.Flags[permName];
                 if (!permBit) {
                     console.warn(`[Relay.js Warn] Invalid permission name used in $onlyPerms: ${permName}`);
                     continue; // Skip invalid permission names
                 }
                 if (!member.permissions.has(permBit)) {
                     missingPerms.push(permName);
                 }
             }

             if (missingPerms.length > 0) {
                 context.stopExecution = true;
                 const finalReply = `${replyMessage} (Missing: ${missingPerms.join(', ')})`;
                  await this.parser.sendErrorReply(new Error(finalReply), context); // Send the custom or default reply
                 return ''; // Stop execution flagged
             }

             return ''; // User has all required permissions, continue execution
         }, 1, Infinity);


        // --- Slash Command Option Getter ---
         register('option', (args, context) => { // Synchronous
             if (!context.interaction?.isChatInputCommand()) {
                 // console.warn("[Relay.js Warn] $option used outside of slash command context.");
                 return '';
             }
             if (args.length !== 1) throw new Error("Usage: $option[optionName]");
             const optionName = String(args[0]).toLowerCase();
             const optionsResolver = context.interaction.options;

             // Try different getters based on potential type (more robust than just .get)
             const value = optionsResolver.get(optionName)?.value ?? // General case
                         optionsResolver.getString(optionName) ??
                         optionsResolver.getInteger(optionName) ??
                         optionsResolver.getNumber(optionName) ??
                         optionsResolver.getBoolean(optionName) ??
                         optionsResolver.getUser(optionName)?.id ?? // Return ID for entity types
                         optionsResolver.getMember(optionName)?.id ?? // Member might not be cached
                         optionsResolver.getChannel(optionName)?.id ??
                         optionsResolver.getRole(optionName)?.id ??
                         optionsResolver.getMentionable(optionName)?.id ??
                         optionsResolver.getAttachment(optionName)?.url; // Return URL for attachments

             return value ?? ''; // Return found value or empty string
         }, 1, 1);


         // --- TODO: Add many more functions ---
          // $addRole, $removeRole, $kick, $ban, $timeout
          // $findUser, $findRole, $findChannel (careful with async/performance)
          // $creationDate, $formatDate, $randomInt
          // $editMessage, $deleteMessage
          // $button, $selectMenu (for creating components - these might just return stringified JSON?)
          // $awaitMessages, $awaitInteraction (more complex flow)
          // $httpRequest (careful with security)
          // $eval (EXTREME caution - security risk)
          // Database functions ($setVar, $getVar, $deleteVar, etc.)
          // Embed functions ($createEmbed, $addField, $setColor...) -> maybe return JSON for use in $reply/$sendMessage
    }
}

module.exports = FunctionManager;